/**
 * Created by John on 11/17/2016.
 * Modified by TFG on 7/21/2020.
 *
 * Common routines for handling RabbitMQ queues
 */

function RabbitHelper(winston, config) {
    var self = this;
    this.rabbit = require("rabbot");
    this.winston = winston;
    var logger = (winston) ? winston.loggers.get('main') : console;
    // it can make a lot of sense to share topology definition across
    // services that will be using the same topology to avoid
    // scenarios where you have race conditions around when
    // exchanges, queues or bindings are in place
    require("../config/RabbitTopology.js")(this.rabbit, null, config)
        .then(function() {
            logger.debug("Rabbit Topology Loaded - config : " + JSON.stringify(config.rabbitConfig, null, 4));
        });

    this.rabbit.on("unreachable", function() {
        logger.info("RABBIT Server Unreachable - configuration used : ");
        logger.info("       host -> " + config.rabbitConfig.host);
        logger.info("       user -> " + config.rabbitConfig.user);
        logger.info("   password -> " + config.rabbitConfig.pass);
        logger.info("       port -> " + config.rabbitConfig.port);
        logger.info("      vhost -> " + config.rabbitConfig.vhost);
        process.exit();
    });
}

/**
 * NOTE: Only used for MOCHA TESTS!!!!!! (for now)
 *
 * @param key
 * @param data
 * @param cb
 */
RabbitHelper.prototype.publishGeneric = function(key, data, cb) {
    var i;
    var self = this;
    var send = function(x, idx, max) {
        var direction = key;
        self.rabbit.publish("topic-fresenius-loader", {
            routingKey: direction,
            type: direction,
            body: x,
            persistent: true
        }).then(function() {
            console.log("published " + key + " message", x);
            var out = {
                msg: x,
                idx: idx,
                total: max
            };
            if (cb) cb(null, out);
        }).catch(function(err){
            if (cb) cb(err, null);
            else throw err;
        });
    };

    if (Array.isArray(data)) {
        for (i = 0; i < data.length; i++) {
            send(data[i], i, data.length);
        }
    } else {
        send(data, 0, 1);
    }

};

RabbitHelper.prototype.publishLoaderSF = function(msg, cb) {
    var self = this;
    var direction = "loader.sf";
    self.rabbit.publish("topic-fresenius-loader", {
        routingKey: direction,
        type: direction,
        body: msg
    }).then(function() {
        self.logger.debug("published Loader SF message");
        if (cb) cb(null, msg);
    }).catch(function(err){
        if (cb) cb(err, null);
        else throw err;
    });
};

RabbitHelper.prototype.publishLoaderMT = function(msg, cb) {
    var self = this;
    var direction = "loader.mt";
    self.rabbit.publish("topic-fresenius-loader", {
        routingKey: direction,
        type: direction,
        body: msg
    }).then(function() {
        self.logger.debug("published Loader MT message");
        if (cb) cb(null, msg);
    }).catch(function(err){
        if (cb) cb(err, null);
        else throw err;
    });
};

RabbitHelper.prototype.publishLoaderPV = function(msg, cb) {
    var self = this;
    var direction = "loader.pv";
    self.rabbit.publish("topic-fresenius-loader", {
        routingKey: direction,
        type: direction,
        body: msg
    }).then(function() {
        self.logger.debug("published Loader PV message");
        if (cb) cb(null, msg);
    }).catch(function(err){
        if (cb) cb(err, null);
        else throw err;
    });
};

RabbitHelper.prototype.publishLoaderCP = function(msg, cb) {
    var self = this;
    var direction = "loader.cp";
    self.rabbit.publish("topic-fresenius-loader", {
        routingKey: direction,
        type: direction,
        body: msg
    }).then(function() {
        self.logger.debug("published Loader CP message");
        if (cb) cb(null, msg);
    }).catch(function(err){
        if (cb) cb(err, null);
        else throw err;
    });
};

RabbitHelper.prototype.publishLoaderCR = function(msg, cb) {
    var self = this;
    var direction = "loader.cr";
    self.rabbit.publish("topic-fresenius-loader", {
        routingKey: direction,
        type: direction,
        body: msg
    }).then(function() {
        self.logger.debug("published Loader CR message");
        if (cb) cb(null, msg);
    }).catch(function(err){
        if (cb) cb(err, null);
        else throw err;
    });
};

RabbitHelper.prototype.cleanUp = function(cb) {
    this.rabbit.shutdown().then(function() {
        if (cb) {
            cb(null, null);
        }
    }).catch(function(err) {
        if (cb) {
            cb(err, null);
        }
        else {
            throw err;
        }
    });
};

module.exports = RabbitHelper;