"use strict";

module.exports = function(sequelize, DataTypes) {
    var CatalogList = require("./catalogList")(sequelize, DataTypes);
    var ItemCatalog = require("./itemCatalog")(sequelize, DataTypes);
    var MasterItem = sequelize.define('MasterItem', {
            id: {
                type: DataTypes.INTEGER,
                field: 'id',
                primaryKey: true,
                autoIncrement: true
            },
            sapMaterial: {
                type: DataTypes.INTEGER,
                field: 'sap_material_id'
            },
            description: {
                type: DataTypes.STRING,
                field: 'material_description'
            },
            cntInd: {
                type: DataTypes.BOOLEAN,
                field: 'cnt_ind',
                default: false
            },
            cntDescription: {
                type: DataTypes.STRING,
                field: 'cnt_description'
            },
            medVendor: {
                type: DataTypes.VIRTUAL
            },
            vendorMaterial: {
                type: DataTypes.STRING,
                field: 'vendor_material'
            },
            conv: {
                type: DataTypes.INTEGER,
                field: 'conv'
            },
            countType: {
                type: DataTypes.STRING,
                field: 'count_type'
            },
            showInventoryInd: {
                type: DataTypes.BOOLEAN,
                field: 'onhand_inventory_ind',
                default: false
            },
            dateModified: {
                type: DataTypes.DATE,
                field: 'date_modified'
            },
            discontinued: {
                type: DataTypes.DATE,
                field: 'discontinued'
            },
            alternates: {
                type: DataTypes.STRING,
                field: 'alternates'
            },
            message: {
                type: DataTypes.STRING,
                field: 'item_message'
            },
            ndc: {
                type: DataTypes.STRING,
                field: 'ndc'
            },
            vendorId: {
                type: DataTypes.STRING,
                field: 'vendor_id'
            },
            unitPrice: {
                type: DataTypes.DECIMAL(9,2),
                field: 'unit_price'
            }
        },
        {
            timestamps: true,
            createdAt: false,
            updatedAt: 'date_modified',
            deletedAt: 'discontinued',
            paranoid: true,
            tableName: 'tbl_master_item',
            indexes: [
                // Create a unique index on device timestamp
                {
                    unique: true,
                    fields: ['sapMaterial','vendorMaterial']
                },
            ],
            getterMethods   : {
                catalogAsString       : function()  {
                    if (this.Catalogs) {
                        if (Array.isArray(this.Catalogs)) {
                            var out = "";

                            for (var r = 0; r < this.Catalogs.length; r++) {
                                var cat = this.Catalogs[r];
                                out += ((r === 0) ? "" : ", ") + cat.catalogId;
                            }

                            return out;
                        }
                        else
                            return this.Catalogs.catalogId
                    } else {
                        return "";
                    }
                }
            }
        }
    );
    MasterItem.belongsToMany(CatalogList, { through: ItemCatalog , as: 'Catalogs', foreignKey: 'itemId', otherKey: 'catalogId'});
    return MasterItem;
};
