/**
 * Created by TFG on 07/15/2020.
 * Management for the Master Item loading data
 *
 * CAUTION ! August 2020 note !
 * Since Clinic Profile process is not complete, in order to have an intermediate release CP will ignore any record
 * coming in the queue !
 * See function processMetadata - set abort flag to true !
 * That flag should be set to false for normal working !
 */
var app = require("../../child");

var LoaderHelper = require("../helpers/LoaderHelper");
var awcblcommon = require('../config/awcblcommon');

var expectedLoadingTime = 120000;     // Expected execution time (watchdog) in ms (default 2 minutes - 120 sec - 120000 ms)
var loadingCPwatchDog = null;         // Global variable to keep the loading timeout timer

var debug = LoaderHelper.debug;

var currentTransaction;
// Doing that because unable to assign correctly the models variable where needed
var clinicCatalog = {
    clinicId: "",      // <--- CAUTION ! This clinic ID is actually the ID on the tbl_clinic_profile
    catalogId: ""
};

// Object to deal with metadata
var CPmetadata = {
    transactionId: "N/A",     //    - transaction ID
    fileName: "--",           //    - name of the originating CSV file
    expectedNumRec: 0,        //    - number to record to be received (remove 1 when counting)
    rcvCounter: 0,            //    - counter of record received  (remember to add 1 when showing the value - counter from 0 not 1)
    dbCounter: 0,             //    - counter of record saved on DB ((remember to add 1 when showing the value - counter from 0 not 1)
    endLoad: false,           //    - flag to indicate the current load as ended
    endProc: false,           //    - flag to indicate the current processing as ended
    abort: false,             //    - flag to abort the transaction ! If true all the incoming records will ack but not processed
    sendEndLoad: true,        //    - flag to be used to send out only ONE log notification about end loading
    sendEndProc: true,        //    - flag to be used to send out only ONE log notification about end processing
    sendAbort: true,          //    - flag to be used to send out only ONE log notification about abort
    incrementRcvCounter : function() {
        this.rcvCounter++;
        if (this.rcvCounter >= (this.expectedNumRec -1)) {
            this.endLoad = true;
        }
    },
    incrementDbCounter : function() {
        this.dbCounter++;
        // Here the test is little bit different because the position of the incrementDbCounter
        // Starting from zero, when process the number 9 (the latest) it will be incremented by one and
        // then the test
        if (this.dbCounter >= this.expectedNumRec) {
            this.endProc = true;
        }
    }
};

// Create array of metadata objects
var metadata = [ CPmetadata ];

module.exports = function(winston, rabbit, models, config, cb) {
    // variable to hold starting time
    var started = Date.now();

    // expected message count
    var logger = winston.loggers.get("main");

    // always setup your message handlers first
    rabbit.nackOnError();

    // this handler will handle messages sent from the publisher
    rabbit.handle({
        queue: 'topic-fresenius-loader-cpq',
        type: 'loader.cp'
    }, function (msg) {
            logger.info(process.pid + ":LOADER CP POST Received:",
            JSON.stringify(msg.body,null, 4),
            "routingKey:", msg.fields.routingKey);
        var helper = new LoaderHelper(models, winston);

        // Metadata management - use the currentTransaction to monitor the load
        currentTransaction = processMetadata(winston, msg.body);

        // Intercept wrong incoming records !
        if (currentTransaction === null ||
            currentTransaction === undefined) {
            // Something wrong handling metadata !!!
            // Ignore incoming record !
            msg.ack();      // Acknowledge the message in the queue to clean it up
        } else {
            if (currentTransaction.abort === true) {
                // Something wrong handling metadata !!!
                // Abort upload !!!

                if (currentTransaction.sendAbort === true) {
                    // Write log to notify successfull end of operation
                    awcblcommon.writeAwcblEntryLog(app,
                        currentTransaction.fileName,
                        currentTransaction.transactionId,
                        "BLOCKED",
                        "Transaction ignored - metadata error").then(response => {
//                        logger.debug(response);
                    }, reject => {
                        logger.error(reject);
                    });
                    currentTransaction.sendAbort = false;     // Disable sending other LOG messages about abort
                }
                msg.ack();
            } else {
                // NOT ABORT !! Process the incoming record !

                // Store the record on the DB via helper
                helper.saveLoaderCP(msg.body).then(response => {
                    // Response - last record saved
                    var mainCPrecordID = response;
                    if (debug === true) {
                        logger.debug("Record " + currentTransaction.dbCounter + " (id " + mainCPrecordID + ") saved ");
                    }

                    currentTransaction.incrementDbCounter();        // Increment saved record counter

                    // Load a timer for about 2 minutes !
                    // If it expires it means that no more records were received and not all the expected records
                    // are arrived
                    // If a previous timer is running, stop it
                    if (loadingCPwatchDog) {
                        if (debug === true) {
                            logger.debug("=======> Loading WatchDog ! Canceled for CP !");
                        }
                        clearTimeout(loadingCPwatchDog);       // Cancel timeout
                    }

                    loadingCPwatchDog = setTimeout(function () {
                        logger.error("CP loading timeout EXPIRED !");

                        // Write log to notify error happened
                        awcblcommon.writeAwcblEntryLog(app,
                            currentTransaction.fileName,
                            currentTransaction.transactionId,
                            "ERROR",
                            "Missing expected records - " + expectedLoadingTime + " ms timer expired - Contact assistance").then(response => {
//                                    logger.debug("AWCBL log written : " + response);
                        }, reject => {
                            logger.error("AWCBL log : " + reject);
                        });
                    }, expectedLoadingTime);

                    // Extract and prepare the associated datat to the record to be saved
                    // The data to be loaded are stored in the array clinicCatalogs
                    var clinicCatalogs = processAssociatedData(currentTransaction.dbCounter, winston, msg.body, mainCPrecordID);

                    // Remove old entries from tbl_clinic_catalog for the current clinic
                    helper.removeCPassociated(mainCPrecordID).then(response => {
                        if (debug === true) {
                            logger.debug("There are " + clinicCatalogs.length + " associated entries of " + mainCPrecordID);
                        }
                        // Removed associated data !!!
                        // Now save the new associated data

                        for (var fn = 0; fn < clinicCatalogs.length; fn++) {
                            // Call here the function to save the associated data
                            helper.saveCPassociatedEntry(clinicCatalogs[fn].clinicId, clinicCatalogs[fn].catalogId).then(response => {
                                // Saved associated data !
                                logger.info("Associated data to clinic record id " + response + " saved");
                            }, reject => {
                                logger.error("Error from helper for CP associated : " + reject);

                                // Write log to notify error happened
                                awcblcommon.writeAwcblEntryLog(app,
                                    currentTransaction.fileName,
                                    currentTransaction.transactionId,
                                    "ERROR",
                                    "Error updating associated : " + reject + " - Contact assistance").then(response => {
//                                    logger.debug(response);
                                }, reject => {
                                    logger.error("AWCBL log : " + reject);
                                });
                            });
                        }
                   }, reject => {
                        logger.error("Error removing for CP associated : " + reject);

                        // Write log to notify error happened
                        awcblcommon.writeAwcblEntryLog(app,
                            currentTransaction.fileName,
                            currentTransaction.transactionId,
                            "ERROR",
                            "Error removing old associated data : " + reject + " - Contact assistance").then(response => {
//                                logger.debug(response);
                        }, reject => {
                            logger.error("AWCBL log : " + reject);
                        });
                    });

                    if (currentTransaction.endProc === true && currentTransaction.sendEndProc === true) {
                        // If a previous timer is running, stop it
                        if (loadingCPwatchDog) {
                            if (debug === true) {
                                logger.debug("=======> Loading WatchDog ! Canceled for PV !");
                            }
                            clearTimeout(loadingCPwatchDog);       // Cancel timeout
                            loadingCPwatchDog = null;
                        }

                        // End of process saving main table !!
                        // This is not YET the end of processing since the tbl_clinic_catalog needs now to be updated !

                        // Write log to notify successfull end of main processing
                        awcblcommon.writeAwcblEntryLog(app,
                            currentTransaction.fileName,
                            currentTransaction.transactionId,
                            "PROCESSED",
                            "Processed " + currentTransaction.dbCounter + "  records").then(response => {
//                            logger.debug(response);
                        }, reject => {
                            logger.error("AWCBL log : " + reject);
                        });

                        currentTransaction.sendEndProc = false;     // Disable sending other LOG messages about ending process
                    }
                    msg.ack();
                }, reject => {
                    // Record not saved ! Ack message anyway to remove from the queue
                    msg.ack();
                    logger.error("Error from helper for CP main : " + reject);

                    // Write log to notify error happened
                    awcblcommon.writeAwcblEntryLog(app,
                        currentTransaction.fileName,
                        currentTransaction.transactionId,
                        "ERROR",
                        "Error loading : " + reject + " - Contact assistance").then(response => {
//                        logger.debug(response);
                    }, reject => {
                        logger.error("AWCBL log : " + reject);
                    });
                });

                // Shows in the log statistics about processing
                logger.info("Transaction " + currentTransaction.transactionId +
                    " - expected " + currentTransaction.expectedNumRec +
                    " - received " + currentTransaction.rcvCounter +
                    " - stored " + currentTransaction.dbCounter +
                    " - end load : " + currentTransaction.endLoad.toString() +
                    " - end proc : " + currentTransaction.endProc.toString());

                if (currentTransaction.endLoad === true && currentTransaction.sendEndLoad === true) {
                    logger.info("Transaction " + currentTransaction.transactionId + " ENDED !");

                    // Write log to notify successfull end of operation
                    awcblcommon.writeAwcblEntryLog(app,
                        currentTransaction.fileName,
                        currentTransaction.transactionId,
                        "LOADED",
                        "Loaded " + (currentTransaction.rcvCounter + 1) + " records").then(response => {
//                        logger.debug(response);
                    }, reject => {
                        logger.error("AWCBL log : " + reject);
                    });
                    currentTransaction.sendEndLoad = false;     // Disable sending other LOG messages about ending loading
                }
            }
        }
    });

    logger.info(process.pid + ':Set up loader cpq OK');
    return require("../config/RabbitTopology.js")(rabbit, "loader.cp", config);
};

// Function to handle the metadata related to a load.
// In case of errors, the function will return an object with the abort flag set to true
function processMetadata( winston, msg) {
    var path = require('path');
    var logger = winston.loggers.get("main");

    var inTransactionId;
    var inFileName;
    var inExpectedNumRec;

    // Extract metadata from input message - it is expected :
    //   csv_line[0] -> transactionId
    //   csv_line[1] -> file name
    //   csv_line[2] -> expected number of messages
    try {
        var csv_line = msg.split(/,(?=(?:(?:[^"]*"){2})*[^"]*$)/);
        csv_line = csv_line || [];

        // Retrieve metadata
        inTransactionId = csv_line[LoaderHelper.rowCPPos.transactionId].split(',').join('');
        inFileName = path.basename(csv_line[LoaderHelper.rowCPPos.fileName].split(',').join(''), ".csv") + ".csv";
        inExpectedNumRec = parseInt(csv_line[LoaderHelper.rowCPPos.expNumber].split(',').join(''), 10);
    } catch (err) {
        logger.error("Process Metadata ! Error ! : " + err);
        // Probably the format of the input is wrong !
        return null;
    }

    // Check if the transaction ID exists on the array
    let currentTransaction = metadata.find(CPmetadata => CPmetadata.transactionId === inTransactionId);

    if (currentTransaction !== undefined) {
        currentTransaction.incrementRcvCounter();
    } else {
        logger.info("New transaction identified : " + inTransactionId);

        // Not found object - add one
        CPmetadata.transactionId    = inTransactionId;
        CPmetadata.fileName         = inFileName;
        CPmetadata.expectedNumRec   = inExpectedNumRec;
        CPmetadata.rcvCounter       = 0;
        CPmetadata.dbCounter        = 0;
        CPmetadata.endLoad          = false;
        CPmetadata.endProc          = false;
        CPmetadata.sendEndLoad      = true;
        CPmetadata.sendEndProc      = true;
        CPmetadata.sendAbort        = true;
        CPmetadata.abort            = false; // TFG !!! Force true to ignore incoming records for CP !!!

        metadata.push(CPmetadata);
        currentTransaction = metadata.find(CPmetadata => CPmetadata.transactionId === inTransactionId);
    }

    return currentTransaction;
}

// Function to handle the CP associated data to store after saving all the data
// Store the data to be sent in the global variable clinicCatalogs.
function processAssociatedData( record_no, winston, msg, inClinicIdId) {
    var path = require('path');
    var logger = winston.loggers.get("main");

    var inClinicId;
    var inTestCatalog;

    var newList = [];

    try {
        var csv_line = msg.split(/,(?=(?:(?:[^"]*"){2})*[^"]*$)/);
        csv_line = csv_line || [];

        // Retrieve catalog related data !!
        inClinicId = csv_line[LoaderHelper.rowCPPos.clinicId].split(',').join('');
        inTestCatalog = csv_line[LoaderHelper.rowCPPos.testCatalog].split(',').join('');
    } catch (err) {
        logger.error("Process Associated data ! Error ! : " + err);
        // Probably the format of the input is wrong !
        return;
    }

    if (debug === true) {
        logger.debug("processAssociatedData - record number = " + record_no);
        logger.debug("                      - clinic ID     = " + inClinicId);
        logger.debug("                      - clinic ID ID  = " + inClinicIdId);
        logger.debug("                      - catalog ID    = " + inTestCatalog);
    }
    // Prepare the list of the record for tbl_clinic_catalog to add
    var catIds = inTestCatalog.split('\#');

    if (debug === true) {
        logger.debug("                      Found " + catIds.length + " catalogs ID");
    }
    for (var fn = 0; fn < catIds.length; fn++) {
//        clinicCatalogs[fn] = new createClinicCatalogEntry(inClinicIdId, catIds[fn]);
        if (debug === true) {
            logger.debug("                        --> associate " + inClinicIdId + " to " +catIds[fn]);
        }
        newList[fn] = new createClinicCatalogEntry(inClinicIdId, catIds[fn]);
    }

    return newList;
}

function createClinicCatalogEntry(clinic_id, catalog_id) {
    this.clinicId = clinic_id;
    this.catalogId = catalog_id;
}
