/**
 * Created by John on 4/4/2017.
 */
var cluster = require('cluster');
var clusterManager = require('cluster-manager');
var port = process.env.PORT || 3000;
var numCPUs = require('os').cpus().length;
var env = process.env.NODE_ENV || "dev";
var path = require('path');

var webpath = path.join(__dirname, 'childWeb.js');
var combinedPath = path.join(__dirname, 'child.js');
var qpath = path.join(__dirname, 'childQueues.js');

var config      = require("./components/config/server." + env + ".js");

//initialize logging
var winston = require('winston');
var win = require('./components/config/logging')(winston, config);
var logger = win.loggers.get("main");

// FRES-209
// If the host is : ip-172-31-136-230.ec2.internal
// the process will be forced to stop immediately
//
//  CAUTION !!!!  If for any reason the IP/Hostanme is changing, it is necessary to update this value !
//
var os = require("os");
var hostname = os.hostname();
var checkHostname = "ip-172-31-136-230.ec2.internal";
//var checkHostname = "ip-172-31-141-174.ec2.internal";  // QA test

if (numCPUs === 1) {
    throw new Error("Use server.js for single core machines!");
}

var singleChildScript = true;

if (cluster.isMaster) {
    var isEven = (numCPUs % 2 === 0);
    if (!isEven && !singleChildScript) numCPUs--;

    // TFG - Creating more forks should helps to spread the load of the queue management on different cpus.
    // For now creating only 1 process. IN case of load problems is possible to restore the setting, however should
    // be beneficial only if different queues are engaged - not sure at this point how a single queue can be affected
    // having the process split on different forks.
//    var forks = (env == "dev") ? 2 : ((singleChildScript) ? numCPUs : numCPUs / 2);
    var forks = 1;
    logger.info("/n-------------------------------------------------------------------");
    logger.info('Environment : ' + env + " Forks : " + forks + " Num cpus : " + numCPUs + " - Hostname " + hostname);

    // FRES-209
    if (hostname.toString().localeCompare(checkHostname) === 0) {
        logger.info("ATTENTION ! On this machine this process need to be kept stopped ! Terminate it NOW !!");
        // TFG - process.exit() is not maybe the best way to gracefully exit BUT in this case should be OK since nothing
        // is actually started yet, so no thread/promises should be around at this stage.
        process.exit(0);
    }

    cluster.on('online', function(worker) {
        logger.info('Worker ' + worker.process.pid + ' is online');
    });

    cluster.on('exit', function(worker, code, signal) {
        if (worker.exitedAfterDisconnect === true) {
            logger.debug('Oh, it was just voluntary – no need to worry');
        }
        if (signal) {
            logger.debug('worker was killed by signal: ' + signal);
        } else if (code !== 0) {
            logger.debug('worker exited with error code: ' + code);
        } else {
            logger.debug('worker success!');
        }
    });

    var realPath = (singleChildScript) ? combinedPath : webpath;

    logger.info( " Starting " + forks + ((singleChildScript) ? " combined: " : " web: ") + realPath);

    clusterManager({
        exec: realPath,
        workers: forks,
        notify: null
    });

    //allocate the extra core to queue drainage
    if (!singleChildScript) {
        if (!isEven) forks++;

        logger.info(" Starting " + forks + " Queues: " + qpath);

        new clusterManager({
            exec: qpath,
            workers: forks,
            notify: null
        });
    }
}

process.on('uncaughtException', function (err) {
    var msg = (new Date).toUTCString() + ' uncaughtException:';
    logger.error(msg, err.message);
    logger.error(err.stack);
    process.exit(1);
});